package classReWriter;

import org.objectweb.asm.*;
import org.objectweb.asm.commons.*;
import java.io.*;
import java.lang.System;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class ETAdapter extends ClassAdapter {
	
  /**
   * the ASM name for this class, such as "java/lang/String"
   */
  private String className;

  /**
   * the unique id assigned to this class
   */
  private int classId;

  /**
   * indicates whether the optimizing method adapter was requested
   */
  private boolean useOptimizer;

  /**
   * flag settings to use if -x but no string given;
   * meanings are:
   * @ a place holder - no flags, but experimental indicated
   * E - emit early MethodEntry call for constructors
   * F - include field ids in PUTFIELD and PUTSTATIC instrumentation calls
   * I - emit inifOf calls
   * L - emit late MethodEntry call for constructors (normally use exactly one of E and L)
   * N - emit newobj/newrefobj calls at NEW bytecodes
   * O - emit newobj call in Object.<init> (normally use exactly one of N and O)
   * R - emit refInit instrumentation calls
   * T - emit MethodExit at ATHROW (normally use exactly one of T and W)
   * U - emit UninitPutfield instrumentation
   * W - add catch-everything wrapper and call the MethodException instrumentation
   */
  private static final String defaultXFlags = "EFINRWU";

  /**
   * flag settings to use if -x is not given
   */
  private static final String defaultNoXFlags = "LOT";

  /**
   * parsed form of the experimental flags passed
   */
  private Set<String> expFlags;

  /**
   * stream for outputting names information
   */
  private PrintStream nameStream;

  /**
   * the last unique class number assigned
   */
  private static int lastClassId = 0;

  /**
   * table giving the unique class number for each class
   */
  static final Map<String,Integer> classIds = new HashMap<String,Integer>();

  /**
   * looks up the unique id of a class, assigning a new one if necessary
   * @param className a String giving the name of the class whose id is desired
   * @return an int giving that class's unique id, which is assigned now if the
   * class has not been seen before
   */
  public static int getClassId (String className) {
    if (className == null) {
      // superclass of Object
      return 0;
    }
    Integer id = classIds.get(className);
    if (id == null) {
      id = ++lastClassId;
      classIds.put(className, id);
    }
    return id;
  }

  /**
   * maps an internal class name, such as "java/lang/String"
   * to a map giving the unique number assigned to each of its fields
   */
  protected static final Map<String, Map<String, Integer>> fieldIdTables =
    new HashMap<String, Map<String, Integer>> ();

  /**
   * last field id assigned
   */
  private static int lastFieldId = 0;

  /**
   * get the unique id of this field, identified by its owning class
   * and its Java field name
   * @param owningClass a String giving the owning class, as in "java/lang/String"
   * @param fieldName a String giving the Java name of the field, as in "foo"
   * @param isStatic a boolean that is true iff this is a static field (currently ignored)
   * @result an int giving the unique number we have assigned to this field
   */
  static int getFieldId (String owningClass, String fieldName, boolean isStatic) {
    Map<String, Integer> fieldIds = fieldIdTables.get(owningClass);
    if (fieldIds == null) {
      fieldIds = new HashMap<String, Integer>();
      fieldIdTables.put(owningClass, fieldIds);
    }
    Integer id = fieldIds.get(fieldName);
    if (id == null) {
      id = ++lastFieldId;
      fieldIds.put(fieldName, id);
    }
    return id;
  }

  /**
   * the immediate sublcasses of java/lang/ref/Reference
   */
  private static Set<String> refSubclasses = new HashSet<String>();

  // the classes we want to track
  static {
    refSubclasses.add("java/lang/ref/SoftReference");
    refSubclasses.add("java/lang/ref/WeakReference");
  }

  /**
   * indicates whether a class is an interesting immediate subclass
   * of java/lang/ref/Reference
   * @param name a String giving the name of the class being looked up
   * @result a boolean that is true iff the class is an interesting
   *         immediate subclass of java/lang/ref/Reference
   */
  public static boolean isReferenceSubclass (String name) {
    return refSubclasses.contains(name);
  }

  /**
   * classes that are, or inherit from, the interesting Reference subclasses
   * NOTE: For proper initialization, this line must come after the initialization
   * of refSubclasses!
   */
  private static Set<String> refInheritors = new HashSet<String>(refSubclasses);
  
  /**
   * indicates whether a class is, or inherits from, an interesting subclass
   * of java/lang/ref/Reference
   * @param name a Strig givine the name of the class being looked up
   * @result a boolean that is the iff the class is an interesting inheritor from
   *         java/lang/ref/Reference
   */
  public static boolean isReferenceInheritor (String name) {
    return refInheritors.contains(name);
  }


  public ETAdapter (ClassVisitor cv, boolean useOptimizer, String xFlags,
                    PrintStream nameStream) {
    super(cv);
    this.useOptimizer = useOptimizer;
    this.nameStream   = nameStream;
    this.expFlags     = prepareXFlags(xFlags);
  }
	
  /**
   * parse experimental flags string
   * @param xFlags a String with the experimental flags as passed
   * @result a Set<String> giving the parsed flags
   */
  private Set<String> prepareXFlags (String xFlags) {
    if (xFlags == null) {
      xFlags = defaultNoXFlags;
    } else if (xFlags.length() == 0) {
      xFlags = defaultXFlags;
    }

    Set<String> flags = new HashSet<String>();

    // For now this is quite simple: each character forms one flag string
    for (int i = 0, n = xFlags.length(); i < n; ++i) {
      flags.add(xFlags.substring(i, i+1));
    }

    return flags;
  }

  @Override
  public void visit (int version, int access, String name, String signature,
                     String superName, String[] interfaces ){
    cv.visit(version, access, name, signature, superName, interfaces);

    className = name;

    if (isReferenceInheritor(superName)) {
      refInheritors.add(name);
    }

    if ((access & Opcodes.ACC_INTERFACE) == 0) {
      // class, not interface
      classId = getClassId(className);
      int superId = getClassId(superName);
      nameStream.printf("C 0x%x %s 0x%x %s%n", classId, className, superId,
                        (isReferenceInheritor(name) ? "N" : "R"));
    } else {
      // interface: no harm in assigning a number
      // helpful to have a line to bracket the method lines
      classId = getClassId(className);
      nameStream.printf("I 0x%x %s%n", classId, className);
    }
    nameStream.flush();

  }

  @Override
  public FieldVisitor visitField (final int access, final String name, final String desc,
                                  final String signature, final Object value) {
    String instance = ((access & Opcodes.ACC_STATIC) == 0) ? "S" : "I";

    int fieldId = getFieldId(className, name, false);

    nameStream.printf("F %s 0x%x %s 0x%x %s %s%n", instance, fieldId, name, classId, className, desc);

    return super.visitField(access, name, desc, signature, value);
  }

  @Override 
  public MethodVisitor visitMethod (int access, String name, String desc,
                                    String signature, String[] exceptions) {
		
    if (skipMethod(access, name, desc, signature, exceptions)) {
      System.err.println("Skipping method name:" +className + "#" + name + " "+ desc);
      return cv.visitMethod(access, name, desc, signature, exceptions);
    }

    ClassReWriter.verbosePrintln("Instrumenting method name: " + name + " " + desc);
    MethodVisitor writer = cv.visitMethod(access, name, desc, signature, exceptions);
    if (useOptimizer) {
      return new ETOptMethodAdapter(access, className, name, desc, writer, expFlags, nameStream);
    } else {
      return new ETMethodAdapter(access, className, name, desc, writer, nameStream);
    }
  }

  @Override
  public void visitEnd () {
    // end-of-class (output for interfaces as well)
    nameStream.printf("E 0x%x %s%n", classId, className);
    cv.visitEnd();
  }

  private boolean skipMethod(int access, String name, String desc,
                             String signature, String[] exceptions){
    return false;
  }

}

// Local Variables:
// mode:Java
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
